/* Licensing: GNU GPL                    */
/* Author: Felipe Bergo <bergo@seul.org> */
/* $Id: xpete.c,v 1.6 2000/12/11 23:30:32 bergo Exp $ */

#define XVERSION "2000.12.10"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../config.h"

#include "xpete.h"
#include "music.h"
#include "mousepad.h"
#include "songs.h"
#include "stages.h"

#include "gx.h"

GtkWidget *mainw, *da;
GdkPixmap *canvas;

GdkPixmap *enemy[3];
GdkPixmap *pete[5];
GdkPixmap *mbox[2];
GdkPixmap *ball[5];
GdkPixmap *boom[7];
GdkPixmap *life[2];
GdkPixmap *rip[16];
GdkPixmap *introp[1];
GdkPixmap *pup[10];

GdkFont *big,*avg,*normal;
int to_tag;

/* ------- GAME GLOBALS -------- */

int tick=0;
int shoot_pace=4;
int lives=3;
int Stage=0;
char StageName[64];
int Score=0;
int death_sequence=0;
int intro=1;
int prelude=0;
int gameover=0;
int hasmusic=0;
int congrats=0;

int laser_w=0;
int laser_g=0;

/* ------- game actors --------- */

struct Enemy {
  int index;
  int x,y;
  int vx,vy;
  int maxx,maxy;
  int minx,miny;
  int w,h;
  int explode;
  int prize;
  int shootness; /* 0-1000, shooting probability */
  int shootmask; /* arcane bit mask for shooting direction */
  int three_day_waiting_period;
};

/* no one dared to write this in C before :)))) */
struct Hero {
  int x,y;
  int lp; /* leg position */
  int w,h;
  int xspeed;
  /* power ups */
  int PU_Speed;
  int PU_TriShot;
  int PU_Shield;
  int PU_Laser;
} Pete;

struct Shot {
  int x,y;
  int w,h;
  int yspeed,xspeed;
  int index;
  int tick;
  int friend; /* 0=foe shot 1=friend shot */
} flying[200];

int nflying=0;

struct PowerUp {
  int index;
  int appear;
  int x;
  int y;
  int w;
  int h;
  int vx;
  int vy;
  int timeout;
} steroids[50];

int nsteroids=0;

int ldown=0,rdown=0;

struct Enemy fodder[64];
int nfodder;

/* cmd line */
int wantmusic=1;
int wantfx=1;
int wantmousepad=0;

/* game data */

static char *sintro[]={"Help Pete explode Iridium, Mir",
		       "UFOs and other things that hang",
		       "up in the sky and could fall on",
		       "your head if he didn't shoot at them"};

static char *scredits[]={"(C) 2000 Felipe Bergo <bergo@seul.org>, Distributed under the terms of the GNU General Public",
			 "License, version 2 or later. Check http://www.gnu.org/copyleft/gpl.html for details."};

int pusher[8]={0,96,192,240,255,240,192,96};

int main(int argc, char **argv) {
  int i;

  for(i=1;i<argc;i++) {
    if ((!strcmp(argv[i],"-h"))||(!strcmp(argv[i],"-?"))) {
      fprintf(stderr,"xpete - command line options:\n\n");
      fprintf(stderr,"-nm      : no music\n");
      fprintf(stderr,"-ns      : no sound fx\n");
      fprintf(stderr,"-rat     : open mouse control window\n");
      fprintf(stderr,"-h / -?  : this help\n\n");
      return 778;
    }
    if (!strcmp(argv[i],"-nm"))
      wantmusic=0;
    if (!strcmp(argv[i],"-ns"))
      wantfx=0;
    if (!strcmp(argv[i],"-rat"))
      wantmousepad=1;
  }

  srand(time(NULL));

  gtk_init(&argc,&argv);
  gdk_rgb_init();

#ifdef HAVE_SYS_SOUNDCARD_H
  if (wantmusic) {
    if (!music_open())
      hasmusic=1;
    else
      fprintf(stderr,"** [xpete] couldn't open /dev/sequencer -- no music :(\n");
  }
#endif
    
  make_game();

  if (wantmousepad)
    create_pad();

  gtk_main();

#ifdef HAVE_SYS_SOUNDCARD_H
  if (wantmusic) {
    if (hasmusic) {
      music_reset();
      music_close();
    }
  }
#endif

  return 0;
}

void make_game() {
  GtkStyle *sty;
  GdkBitmap *mask;

  init_stage(0);

  mainw=gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_events(mainw,GDK_KEY_PRESS_MASK|GDK_KEY_RELEASE_MASK);
  gtk_window_set_title(GTK_WINDOW(mainw),"X Pete");
  gtk_widget_set_usize(mainw,640+96,480+10);
  
  da=gtk_drawing_area_new();
  gtk_widget_set_events(da,GDK_EXPOSURE_MASK);
  gtk_drawing_area_size(GTK_DRAWING_AREA(da),640+96,480+10);
  gtk_container_add(GTK_CONTAINER(mainw),da);

  gtk_widget_realize(mainw);
  gtk_widget_realize(da);
  sty=gtk_widget_get_style(mainw);

  normal=gdk_font_load("-*-helvetica-medium-r-*-*-12-*-*-*-*-*-*-*");
  big=gdk_font_load("-*-helvetica-bold-r-*-*-24-*-*-*-*-*-*-*");
  avg=gdk_font_load("-*-helvetica-bold-r-*-*-18-*-*-*-*-*-*-*");


  enemy[0]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,enemy1_xpm);
  enemy[1]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,enemy2_xpm);
  enemy[2]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,enemy3_xpm);
  pete[0]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,pete1_xpm);
  pete[1]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,pete2_xpm);
  pete[2]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,pete3_xpm);
  pete[3]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,pete4_xpm);
  pete[4]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,pete5_xpm);

  mbox[0]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,mailbox1_xpm);
  mbox[1]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,mailbox2_xpm);

  ball[0]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,ball1_xpm);
  ball[1]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,ball2_xpm);

  boom[0]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,boom1_xpm);
  boom[1]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,boom2_xpm);
  boom[2]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,boom1_xpm);
  boom[3]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,boom2_xpm);
  boom[4]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,boom3_xpm);
  boom[5]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,boom4_xpm);
  boom[6]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,boom5_xpm);

  life[0]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,life_xpm);
  life[1]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,dead_xpm);

  rip[0]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,void1_xpm);
  rip[1]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,void2_xpm);
  rip[2]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,void3_xpm);
  rip[3]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,void4_xpm);
  rip[4]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,void5_xpm);
  rip[5]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,void6_xpm);
  rip[6]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,void7_xpm);
  rip[7]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,void8_xpm);
  rip[8]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,void9_xpm);
  rip[9]=rip[7];
  rip[10]=rip[8];
  rip[11]=rip[7];
  rip[12]=rip[8];
  rip[13]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,void10_xpm);
  rip[14]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,void11_xpm);
  rip[15]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,gameover_xpm);

  introp[0]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,intro_xpm);

  pup[9]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,para_xpm);
  pup[0]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,pu1_xpm);
  pup[1]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,pu2_xpm);
  pup[2]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,pu3_xpm);
  pup[3]=gdk_pixmap_create_from_xpm_d(mainw->window,&mask,NULL,pu4_xpm);

  canvas=gdk_pixmap_new(mainw->window,640+96,480+10,-1);

  draw_scene();

  gtk_signal_connect(GTK_OBJECT(da),"expose_event",
                     GTK_SIGNAL_FUNC(exposed),NULL);

  gtk_signal_connect(GTK_OBJECT(mainw),"destroy",
                     GTK_SIGNAL_FUNC(wdestroy),NULL);

  gtk_signal_connect(GTK_OBJECT(mainw),"delete_event",
                     GTK_SIGNAL_FUNC(wdelete),NULL);


  gtk_signal_connect(GTK_OBJECT(mainw),"key_press_event",
		     GTK_SIGNAL_FUNC(keydown),NULL);
  gtk_signal_connect(GTK_OBJECT(mainw),"key_release_event",
		     GTK_SIGNAL_FUNC(keyup),NULL);

  gtk_widget_show(da);
  gtk_widget_show(mainw);

  to_tag=gtk_timeout_add(100,step,NULL);

}

gboolean wdestroy(GtkWidget *w,GdkEvent *ev,gpointer data) {
  gdk_font_unref(normal);
  gdk_font_unref(big);
  gtk_main_quit();
}

gboolean wdelete(GtkWidget *w,gpointer data) {
  return FALSE;
}

gboolean exposed(GtkWidget *w,GdkEventExpose *gee,gpointer data)
{
  gdk_draw_pixmap(w->window,
                  w->style->fg_gc[GTK_WIDGET_STATE (w)],
                  canvas,
                  gee->area.x, gee->area.y,
                  gee->area.x, gee->area.y,
                  gee->area.width, gee->area.height);
  return FALSE;
}

void draw_scene() {
  GdkGC *mygc;
  int i,vi,x,y,r,t,c;
  static int intick=0;
  struct Enemy *f;
  struct Shot *s;
  char z[32];

  mygc=gdk_gc_new(da->window);

  gdk_rgb_gc_set_foreground(mygc,0);
  gdk_draw_rectangle(canvas,mygc,TRUE,0,0,640,480);

  if ((prelude)&&(!intro)) {
    draw_prelude(Stage,mygc);
  } else
    if (intro) {
      ++intick;
      gdk_draw_pixmap(canvas,mygc,introp[0],
		      0,0,80,20,475,89);

      gdk_draw_pixmap(canvas,mygc,pete[3],
		      0,0,80,110,96,96);

      gdk_rgb_gc_set_foreground(mygc,0x2aaafc);
      for(i=0;i<4;i++)
	gdk_draw_string(canvas,avg,mygc,180,130+i*22,sintro[i]);

      i=intick%8;
      i = (pusher[i]<<16)|( ((pusher[i]<<1)/3)<<8);
      gdk_rgb_gc_set_foreground(mygc,i);
      gdk_draw_string(canvas,avg,mygc,80,300,"Push Space to Start");

      gdk_rgb_gc_set_foreground(mygc,0xffffff);
      gdk_draw_string(canvas,avg,mygc,10,430,"http://xpete.seul.org      http://www.itsallpetesfault.org");

      gdk_rgb_gc_set_foreground(mygc,0x2aaafc);
      for(i=0;i<2;i++)
	gdk_draw_string(canvas,normal,mygc,10,450+i*14,scredits[i]);
      sprintf(z,"Version: %s",XVERSION);
      gdk_draw_string(canvas,normal,mygc,10,450+2*14,z);

    } else {
      for(i=0;i<nfodder;i++) {
	f=&fodder[i];
	
	if (f->explode) {
	  if (f->explode >= 7)
	    continue;
	  gdk_draw_pixmap(canvas,mygc,boom[f->explode],
			  0,0,f->x+(f->w-48)/2,f->y+(f->h-48)/2,48,48);
	} else
	  gdk_draw_pixmap(canvas,mygc,enemy[f->index],
			  0,0,f->x,f->y,f->w,f->h);  
      }

      if ((laser_g)||(laser_w)) {
	for(i=-laser_w;i<laser_w;i++) {	  
	  c=255;
	  r=(i<0)?(20-8*i):(20+8*i);
	  c<<=16;
	  c|=(r<<8)|r;
	  gdk_rgb_gc_set_foreground(mygc,c);
	  t=Pete.x+70+i;
	  gdk_draw_line(canvas,mygc,t,0,t,480);	
	}
      }
      
      if (gameover) {
	gdk_draw_pixmap(canvas,mygc,rip[15],0,0,193,204,
			254,71);
      } else {
	if (death_sequence)
	  gdk_draw_pixmap(canvas,mygc,rip[death_sequence],0,0,Pete.x,Pete.y,
			  Pete.w,Pete.h);
	else
	  gdk_draw_pixmap(canvas,mygc,pete[Pete.lp],0,0,Pete.x,Pete.y,
			  Pete.w,Pete.h);
      }

      for(i=0;i<nsteroids;i++) {
	if (steroids[i].appear <= tick) {
	  gdk_draw_pixmap(canvas,mygc,pup[9],
			  0,0,steroids[i].x,steroids[i].y,
			  steroids[i].w,steroids[i].h/2);
	  gdk_draw_pixmap(canvas,mygc,pup[steroids[i].index],
			  0,3,steroids[i].x,steroids[i].y+32,
			  steroids[i].w,32-5);
	}
      }
      
      for(i=0;i<nflying;i++) {
	s=&flying[i];
	if (!out_of_screen(i))
	  gdk_draw_pixmap(canvas,mygc,ball[s->index],
			  0,0,s->x,s->y,s->w,s->h);
      }

      /* power up indicators */
      if (Pete.PU_Speed) {
	if ((Pete.PU_Speed>20)||((tick>>1)%2))
	  gdk_draw_pixmap(canvas,mygc,pup[0],0,0,0,480-96-32,48,32);
	gdk_rgb_gc_set_foreground(mygc,0xff0000);
	gdk_draw_rectangle(canvas,mygc,TRUE,48,480-96-32+14,
			   Pete.PU_Speed>>1,4);	  
      }
      if (Pete.PU_TriShot) {
	if ((Pete.PU_TriShot>20)||((tick>>1)%2))
	  gdk_draw_pixmap(canvas,mygc,pup[1],0,0,0,480-96-64,48,32);
	gdk_rgb_gc_set_foreground(mygc,0xff00ff);
	gdk_draw_rectangle(canvas,mygc,TRUE,48,480-96-64+14,
			   Pete.PU_TriShot>>1,4);	  
      }
      if (Pete.PU_Shield) {
	if ((Pete.PU_Shield>20)||((tick>>1)%2))
	  gdk_draw_pixmap(canvas,mygc,pup[2],0,0,0,480-96-96,48,32);
	gdk_rgb_gc_set_foreground(mygc,0x0000ff);
	gdk_draw_rectangle(canvas,mygc,TRUE,48,480-96-96+14,
			   Pete.PU_Shield>>1,4);	  
      }
      if (Pete.PU_Laser) {
	if ((Pete.PU_Laser>20)||((tick>>1)%2))
	  gdk_draw_pixmap(canvas,mygc,pup[3],0,0,0,480-96-128,48,32);
	gdk_rgb_gc_set_foreground(mygc,0x00ffff);
	gdk_draw_rectangle(canvas,mygc,TRUE,48,480-96-128+14,
			   Pete.PU_Laser>>1,4);	  
      }

      if (congrats) {
	if (congrats>20)
	  strcpy(z,"Stage Clear");
	else
	  strcpy(z,"Get Ready");
	vi=gdk_string_width(big,z);
	x=(640-vi)/2;
	y=250;
	gdk_rgb_gc_set_foreground(mygc,0xfff000);
	gdk_draw_string(canvas,big,mygc,x,y,z);
      }
      
      /** score region **/
    
      gdk_rgb_gc_set_foreground(mygc,0x202060);
      gdk_draw_rectangle(canvas,mygc,TRUE,640,0,96,490);
      gdk_draw_rectangle(canvas,mygc,TRUE,0,480,640,10);
      
      
      gdk_rgb_gc_set_foreground(mygc,0xffffff);
      gdk_draw_string(canvas,normal,mygc,650,20,"SCORE");
      gdk_draw_string(canvas,normal,mygc,650,100,"STAGE");
      gdk_draw_string(canvas,normal,mygc,650,200,"LIVES");
      
      gdk_rgb_gc_set_foreground(mygc,0xffff00);
      gdk_draw_string(canvas,normal,mygc,650,430,"'R' Restart");
      gdk_draw_string(canvas,normal,mygc,650,450,"'D' Die");
      gdk_draw_string(canvas,normal,mygc,650,470,"'Q' Quit");
      
      gdk_rgb_gc_set_foreground(mygc,0x6060e0);
      sprintf(z,"%d",Score);
      gdk_draw_string(canvas,big,mygc,650,60,z);
      sprintf(z,"%d",Stage);
      gdk_draw_string(canvas,big,mygc,650,140,z);
      gdk_draw_string(canvas,normal,mygc,650,160,StageName);
      
      for(i=0;i<3;i++)
	gdk_draw_pixmap(canvas,mygc,life[(i<lives)?0:1],
			0,0,640+15,220+(2-i)*60,67,56);
    }
    
  gtk_widget_queue_resize(da);
  gdk_gc_destroy(mygc);
}

void init_stage(int stage) {
  int i;
  int x,y;
  char *p,lc[4096];
  static char *stgsep=" ,";

  Stage=stage+1;

  Pete.w=96;
  Pete.h=96;
  Pete.x=320-Pete.w/2;
  Pete.y=480-Pete.h;
  Pete.lp=0;
  Pete.xspeed=10;
  nflying=0; /* kudos to arma for this */
  prelude=0;
  congrats=0;

#ifdef HAVE_SYS_SOUNDCARD_H
  if ((wantmusic)&&(hasmusic))
    music_reset();
#endif

  switch(stage) {
  case 0: /* stage 1 */
    prelude=1;
    strcpy(StageName,"Iridium");
    load_stage_data(stage);
    Pete.PU_Speed=0;
    Pete.PU_TriShot=0;
    Pete.PU_Shield=0;
    Pete.PU_Laser=0;
    break;
  case 1: /* stages 2, 3 */
  case 2:
    strcpy(StageName,"Iridium");
    load_stage_data(stage);
    break;
  default:
    strcpy(StageName,"Not Done Yet");
    load_stage_data(1);
    break;
  }
}

/* see stages.h */
void load_stage_data(int stg) {
  int i;
  char *p,lc[4096];
  static char *stgsep=" ,";
  strcpy(lc,StageData[stg]);
  p=strtok(lc,stgsep);
  nfodder=atoi(p);
  for(i=0;i<nfodder;i++) {
    fodder[i].index=atoi(strtok(0,stgsep));
    fodder[i].x=fodder[i].maxx=fodder[i].minx=atoi(strtok(0,stgsep));
    fodder[i].y=fodder[i].maxy=fodder[i].miny=atoi(strtok(0,stgsep));
    fodder[i].vx=atoi(strtok(0,stgsep));
    fodder[i].vy=atoi(strtok(0,stgsep));
    fodder[i].minx+=atoi(strtok(0,stgsep));
    fodder[i].miny+=atoi(strtok(0,stgsep));
    fodder[i].maxx+=atoi(strtok(0,stgsep));
    fodder[i].maxy+=atoi(strtok(0,stgsep));
    fodder[i].w=atoi(strtok(0,stgsep));
    fodder[i].h=atoi(strtok(0,stgsep));
    fodder[i].prize=atoi(strtok(0,stgsep));
    fodder[i].shootness=atoi(strtok(0,stgsep));
    fodder[i].shootmask=atoi(strtok(0,stgsep));
    fodder[i].explode=0;
    fodder[i].three_day_waiting_period=0;
  }
  /* power up info */
  strcpy(lc,PowerUpData[stg]);
  p=strtok(lc,stgsep);
  nsteroids=atoi(p);
  for(i=0;i<nsteroids;i++) {
    steroids[i].index=atoi(strtok(0,stgsep));
    steroids[i].appear=atoi(strtok(0,stgsep))+tick;
    steroids[i].x=atoi(strtok(0,stgsep));
    steroids[i].vx=atoi(strtok(0,stgsep));
    steroids[i].vy=atoi(strtok(0,stgsep));
    steroids[i].timeout=atoi(strtok(0,stgsep));    
    steroids[i].w=48;
    steroids[i].h=64-5;
    steroids[i].y=-steroids[i].h;
  }
}

int Lemmings=0;

void motion_step() {
  int i,j,x,y,es;

  if ((intro)||(prelude))
    return;

  if (Lemmings==6) {
    for(i=0;i<nfodder;i++)
      fodder[i].explode=1;
    Lemmings=0;
  }

  if ((nfodder==0)&&(!congrats)) {
    congrats=40;
    nflying=0;
  }

  if (congrats) {
    --congrats;
    if (!congrats)
      init_stage(Stage);
  }

  ++tick;

  if ((laser_g)||(laser_w)) {
    if (laser_g) {
      laser_w+=3;
      if (laser_w>16)
	laser_g=0;
    } else
      laser_w-=3;

    if (laser_w<0) laser_w=0;
  }

  if (death_sequence)
    ++death_sequence;

  if (death_sequence>=15) {
    death_sequence=0;
    --lives;
    Pete.PU_Speed=0;
    Pete.PU_Shield=0;
    Pete.PU_TriShot=0;
    Pete.PU_Laser=0;
    if (!lives)
      gameover=1;
  }

  /* Hero */

  if (Pete.PU_Speed)   --Pete.PU_Speed;
  if (Pete.PU_TriShot) --Pete.PU_TriShot;
  if (Pete.PU_Shield)  --Pete.PU_Shield;
  if (Pete.PU_Laser)   --Pete.PU_Laser;

  if (!gameover) {
    Pete.lp=tick%2;
    es=Pete.xspeed;
    es*=(Pete.PU_Speed)?3:1;
    if (ldown) Pete.x-=es;
    if (rdown) Pete.x+=es;
    if (Pete.x<0) Pete.x=0;
    if (Pete.x>(640-Pete.h)) Pete.x=640-Pete.h;
  }

  /* Enemies */
  for(i=0;i<nfodder;i++) {
    if (fodder[i].explode)
      ++fodder[i].explode;

    if (fodder[i].explode>=7) {
      shift_fodders(i);
      continue;
    }

    if ( fodder[i].y > 550 ) {
      shift_fodders(i);
      continue;
    }

    x=fodder[i].x+=fodder[i].vx;
    if (!(tick%4)) {
      y=fodder[i].y+=fodder[i].vy;
      if ((y<fodder[i].miny)||((y+fodder[i].h)>fodder[i].maxy))
	fodder[i].vy=-fodder[i].vy;
    }

    if ((x<fodder[i].minx)||((x+fodder[i].w)>fodder[i].maxx))
      fodder[i].vx=-fodder[i].vx;

    /* shooting */
    if ((fodder[i].shootness)&&(!fodder[i].three_day_waiting_period)) {
      j=1+(int) (1000.0*rand()/(RAND_MAX+1.0));
      if (j<fodder[i].shootness)
	foe_shoots(i);
    }

    if (fodder[i].three_day_waiting_period)
      fodder[i].three_day_waiting_period--;
  }

  /* power ups */
  for(i=0;i<nsteroids;i++) {
    if (steroids[i].appear < tick) {
      steroids[i].x+=steroids[i].vx;
      steroids[i].y+=steroids[i].vy;
      if (steroids[i].y > 520)
	shift_steroids(i--);
    }
  }

  /* gunfire */
  for(i=0;i<nflying;i++) {
    flying[i].x+=flying[i].xspeed;
    flying[i].y+=flying[i].yspeed;
  }

  while ( (nflying) && (out_of_screen(0)) )
    shift_flyers(0);

  if (!gameover)
    check_hit();
}

void shift_flyers(int v) {
  int i;
  --nflying;
  for(i=v;i<nflying;i++)
    memcpy(&flying[i],&flying[i+1],sizeof(struct Shot));
}

void shift_fodders(int v) {
  int i;
  --nfodder;
  for(i=v;i<nfodder;i++)
    memcpy(&fodder[i],&fodder[i+1],sizeof(struct Enemy));
}

void shift_steroids(int v) {
  int i;
  --nsteroids;
  for(i=v;i<nsteroids;i++)
    memcpy(&steroids[i],&steroids[i+1],sizeof(struct PowerUp));
}

int out_of_screen(int idx) {
  struct Shot *s;
  s=&flying[idx];
  return( (s->x > 640) || 
	  ((s->x+s->w) < 0) ||
	  (s->y > 480) ||
	  ((s->y+s->h) < 0) );
}


gint step(gpointer data) {
#ifdef HAVE_SYS_SOUNDCARD_H
  if ((!intro)&&(!prelude)&&(hasmusic)&&(congrats==25))
    stage_music_advance_riff();
  if ((!intro)&&(!prelude)&&(hasmusic)&&(!(tick%25))&&(!congrats))
    stage_music_play(Stage);
#endif
  motion_step();

  draw_scene();
  return TRUE;
}

gboolean keydown(GtkWidget *widget,
		 GdkEventKey *event,
		 gpointer user_data) {

  switch(event->keyval) {
  case GDK_Left:
  case GDK_KP_Left:
    ldown=1;
    break;
  case GDK_Right:
  case GDK_KP_Right:
    rdown=1;
    break;
  case GDK_KP_Space:
  case GDK_space:
    if ((prelude)&&(!intro)) { prelude=0; break; }
    if ((!gameover)&&(!death_sequence)&&(!intro))
      try_shoot();
    if ((gameover)||(intro))
      new_game();
    break;
  case GDK_q:
  case GDK_Q:
    quit_game();
    break;
  case GDK_r:
  case GDK_R:
    new_game();
    prelude=0;
    break;
  case GDK_d:
  case GDK_D:
    if ((!intro)&&(!prelude)&&(!gameover)&&(!death_sequence))
      death_sequence=1;
    break;
  case GDK_K:
  case GDK_k:
    Lemmings=(Lemmings==0)?1:0;
    return FALSE;
  case GDK_A:
  case GDK_a:
    Lemmings=(Lemmings==1)?2:0;
    return FALSE;
  case GDK_B:
  case GDK_b:
    Lemmings=(Lemmings==2)?3:0;
    return FALSE;
  case GDK_O:
  case GDK_o:
    if (Lemmings==2) {
      Pete.PU_Speed=600;
      Pete.PU_Shield=600;
      Pete.PU_TriShot=600;
      Pete.PU_Laser=600;
    }
    Lemmings=((Lemmings==3)||(Lemmings==4))?(Lemmings+1):0;
    return FALSE;
  case GDK_M:
  case GDK_m:
    Lemmings=(Lemmings==5)?6:0;
    return FALSE;    
  }
  Lemmings=0;
  return FALSE;
}

gboolean keyup(GtkWidget *widget,
	       GdkEventKey *event,
	       gpointer user_data) {

  switch(event->keyval) {
  case GDK_Left:
  case GDK_KP_Left:
    ldown=0;
    break;
  case GDK_Right:
  case GDK_KP_Right:
    rdown=0;
    break;
  }
  return FALSE;
}

void foe_shoots(int nf) {
  /* here goes the magic bitmask: 0x01 -> downward fast
                                  0x02 -> downward slow
                                  0x04 -> lefty fast
                                  0x08 -> righty fast
  */

  fodder[nf].three_day_waiting_period=10;

  if (fodder[nf].shootmask&0x01) {
    flying[nflying].x=fodder[nf].x+fodder[nf].w/2;
    flying[nflying].y=fodder[nf].y+fodder[nf].h;
    flying[nflying].w=10;
    flying[nflying].h=10;
    flying[nflying].xspeed=0;
    flying[nflying].yspeed=10;
    flying[nflying].tick=(nflying)?flying[nflying-1].tick:0;
    flying[nflying].index=1;
    flying[nflying].friend=0;
    ++nflying;
  }
  if (fodder[nf].shootmask&0x02) {
    flying[nflying].x=fodder[nf].x+fodder[nf].w/2;
    flying[nflying].y=fodder[nf].y+fodder[nf].h;
    flying[nflying].w=10;
    flying[nflying].h=10;
    flying[nflying].xspeed=0;
    flying[nflying].yspeed=6;
    flying[nflying].tick=(nflying)?flying[nflying-1].tick:0;
    flying[nflying].index=1;
    flying[nflying].friend=0;
    ++nflying;
  }
  if (fodder[nf].shootmask&0x04) {
    flying[nflying].x=fodder[nf].x+fodder[nf].w/2;
    flying[nflying].y=fodder[nf].y+fodder[nf].h;
    flying[nflying].w=10;
    flying[nflying].h=10;
    flying[nflying].xspeed=-3;
    flying[nflying].yspeed=10;
    flying[nflying].tick=(nflying)?flying[nflying-1].tick:0;
    flying[nflying].friend=0;
    flying[nflying].index=1;
    ++nflying;
  }
  if (fodder[nf].shootmask&0x08) {
    flying[nflying].x=fodder[nf].x+fodder[nf].w/2;
    flying[nflying].y=fodder[nf].y+fodder[nf].h;
    flying[nflying].w=10;
    flying[nflying].h=10;
    flying[nflying].xspeed=3;
    flying[nflying].yspeed=10;
    flying[nflying].tick=(nflying)?flying[nflying-1].tick:0;
    flying[nflying].friend=0;
    flying[nflying].index=1;
    ++nflying;
  }
}

void try_shoot() {
  if ((nflying>0) && ((tick-flying[nflying-1].tick)<shoot_pace))
    return; /* not so fast, cowboy! */

  if (Pete.PU_Laser) { /* Bwa hahahaha */
    if ((laser_w)||(laser_g))
      return;
    laser_w=0;
    laser_g=1;
    return;
  }

  flying[nflying].x=Pete.x+63;
  flying[nflying].y=Pete.y-30+10;
  flying[nflying].w=13;
  flying[nflying].h=30;
  flying[nflying].xspeed=2*(rdown-ldown);
  flying[nflying].yspeed=-10;
  flying[nflying].tick=tick;
  flying[nflying].friend=1;
  flying[nflying].index=0;
  ++nflying;

  if (Pete.PU_TriShot) {
    memcpy(&flying[nflying],&flying[nflying-1],sizeof(struct Shot));
    memcpy(&flying[nflying+1],&flying[nflying-1],sizeof(struct Shot));
    flying[nflying].x+=20;
    flying[nflying].xspeed+=2;
    flying[nflying+1].x-=20;
    flying[nflying+1].xspeed-=2;
    nflying+=2;    
  }
}

void check_hit() {
  int i,j;
  int ax[2],bx[2],ay[2],by[2];

  /* friendly fire <-> enemy fire */

  for(i=0;i<nflying;i++) {
    if (!flying[i].friend)
      continue;
    ax[0]=flying[i].x;
    ax[1]=flying[i].x+flying[i].w;
    ay[0]=flying[i].y;
    ay[1]=flying[i].y+flying[i].h;

    for(j=0;j<nflying;j++) {
      if (flying[j].friend)
	continue;
      bx[0]=flying[j].x;
      bx[1]=flying[j].x+flying[j].w;
      by[0]=flying[j].y;
      by[1]=flying[j].y+flying[j].h;

      if (rects_overlap(ax,ay,bx,by)) {
	flying[i].index=-1;
	flying[j].index=-1;
      }
    }
  }

  for(i=0;i<nflying;i++)
    if (flying[i].index<0)
      shift_flyers(i--);

  /* shot <-> enemy collisions */

  for(i=0;i<nflying;i++) {
    if (!flying[i].friend)
      continue;
    ax[0]=flying[i].x;
    ax[1]=flying[i].x+flying[i].w;
    ay[0]=flying[i].y;
    ay[1]=flying[i].y+flying[i].h;
    for(j=0;j<nfodder;j++) {
      bx[0]=fodder[j].x;
      bx[1]=fodder[j].x+fodder[j].w;
      by[0]=fodder[j].y;
      by[1]=fodder[j].y+fodder[j].h;

      if (rects_overlap(ax,ay,bx,by)) {
	shift_flyers(i);
	if (!fodder[j].explode)
	  Score+=fodder[j].prize;
	fodder[j].explode=1;
      }
    }
  }

  /* laser vs. everything */

  if (laser_w) {
    ax[0]=Pete.x+70-laser_w-10;
    ax[1]=Pete.x+70+laser_w+10;
    ay[0]=0;
    ay[1]=480;
    /* enemies */
    for(j=0;j<nfodder;j++) {
      bx[0]=fodder[j].x;
      bx[1]=fodder[j].x+fodder[j].w;
      by[0]=fodder[j].y;
      by[1]=fodder[j].y+fodder[j].h;

      if (rects_overlap(ax,ay,bx,by)) {
	if (!fodder[j].explode)
	  Score+=fodder[j].prize;
	fodder[j].explode=1;
      }
    }
    /* enemy shots */
    for(j=0;j<nflying;j++) {
      if (flying[j].friend)
	continue;
      bx[0]=flying[j].x;
      bx[1]=flying[j].x+flying[j].w;
      by[0]=flying[j].y;
      by[1]=flying[j].y+flying[j].h;
      
      if (rects_overlap(ax,ay,bx,by))
	flying[j].index=-1;
    }
    for(i=0;i<nflying;i++)
      if (flying[i].index<0)
	shift_flyers(i--);    
  }


  /* enemy fire <-> pete */

  bx[0]=Pete.x;
  bx[1]=Pete.x+Pete.w;
  by[0]=Pete.y;
  by[1]=480;
  for(i=0;i<nflying;i++) {
    if (flying[i].friend)
      continue;
    ax[0]=flying[i].x;
    ax[1]=flying[i].x+flying[i].w;
    ay[0]=flying[i].y;
    ay[1]=flying[i].y+flying[i].h;

    if (rects_overlap(ax,ay,bx,by)) {
      shift_flyers(i);
      if (!Pete.PU_Shield)
	death_sequence=1;
    }
  }

  /* enemy <-> pete collisions */

  if (!death_sequence) {
    ax[0]=Pete.x;
    ax[1]=Pete.x+Pete.w;
    ay[0]=Pete.y;
    ay[1]=480;

    for(j=0;j<nfodder;j++) {
      bx[0]=fodder[j].x;
      bx[1]=fodder[j].x+fodder[j].w;
      by[0]=fodder[j].y;
      by[1]=fodder[j].y+fodder[j].h;
      if (rects_overlap(ax,ay,bx,by)) {
	if (!fodder[j].explode)
	  Score+=fodder[j].prize;
	fodder[j].explode=1;
	if (!Pete.PU_Shield)
	  death_sequence=1;
      }
    }
  }

  /* power ups <-> pete */
  ax[0]=Pete.x;
  ax[1]=Pete.x+Pete.w;
  ay[0]=Pete.y;
  ay[1]=480;
  for(i=0;i<nsteroids;i++) {
    if (steroids[i].appear < tick) {
      bx[0]=steroids[i].x;
      by[0]=steroids[i].y;
      bx[1]=steroids[i].x+steroids[i].w;
      by[1]=steroids[i].y+steroids[i].h;
      if (rects_overlap(ax,ay,bx,by)) {
	switch(steroids[i].index) {
	case 0:	  
	  Pete.PU_Speed+=steroids[i].timeout;
	  break;
	case 1:
	  Pete.PU_TriShot+=steroids[i].timeout;
	  break;
	case 2:
	  Pete.PU_Shield+=steroids[i].timeout;
	  break;
	case 3:
	  Pete.PU_Laser+=steroids[i].timeout;
	  break;
	}
	shift_steroids(i--);
      }     
    }
  }

}

int inrect(int x,int y,int a,int c,int b,int d) {
  return ( (x>=a)&&(x<=b)&&(y>=c)&&(y<=d) );
}

int rects_overlap(int *ax,int *ay,int *bx,int *by) {
  return(inrect(bx[0],by[0],ax[0],ay[0],ax[1],ay[1])||
	 inrect(bx[1],by[0],ax[0],ay[0],ax[1],ay[1])||
	 inrect(bx[1],by[1],ax[0],ay[0],ax[1],ay[1])||
	 inrect(bx[0],by[1],ax[0],ay[0],ax[1],ay[1])||
	 inrect(ax[0],ay[0],bx[0],by[0],bx[1],by[1])||
	 inrect(ax[0],ay[1],bx[0],by[0],bx[1],by[1])||
	 inrect(ax[1],ay[1],bx[0],by[0],bx[1],by[1])||
	 inrect(ax[1],ay[0],bx[0],by[0],bx[1],by[1]));
}

void new_game() {
  intro=0;
  lives=3;
  Score=0;
  gameover=0;
  death_sequence=0;
  init_stage(0);
#ifdef HAVE_SYS_SOUNDCARD_H
  if (hasmusic)
    stage_music_reset();
#endif
}

void quit_game() {
  gdk_font_unref(normal);
  gdk_font_unref(big);
  gtk_main_quit();
}


/* preludes */

int pstep[6]={0,0,0,0,0,0};
int s1rstep=3;

char *lytaletter="Pete, # @ @ "\
"I've just read that some Iridium satellites @ "\
"are about to start falling near the lake. @ # "\
"I read it on the Internet, # # so it must be true! @ @ # # "\
"Let's cancel this weekend's fishing, # ok ? @ @ "\
"Regards -- # Lyta. # # # # # # # # #";

void draw_prelude(int stage,GdkGC *mygc) {
  static int px,py,jc;
  int i,j,lx,ly;
  char bp[512],*p;

  gdk_rgb_gc_set_foreground(mygc,0);
  gdk_draw_rectangle(canvas,mygc,TRUE,0,0,640+96,480+10);

  for(i=0;i<256;i++) {
    gdk_rgb_gc_set_foreground(mygc,255-i);
    gdk_draw_line(canvas,mygc,0,i,640+96,i);
  }

  gdk_rgb_gc_set_foreground(mygc,0x008000);
  gdk_draw_rectangle(canvas,mygc,TRUE,0,480,640+96,10);
 
  switch(stage) {
  case 1:
    switch(pstep[1]) {
    case 0:
      px=-14; py=480-96;
      pstep[1]++;
    case 1: /* pete walking left to right, letter in mailbox */
      px+=14;
      gdk_draw_pixmap(canvas,mygc,pete[2+(px/14)%2],0,0,px,py,96,96);
      gdk_draw_pixmap(canvas,mygc,mbox[0],0,0,500,480-96,96,96);
      s1rstep=3;
      if (px>=400)
	pstep[1]++;
      break;
    case 2: /* letter off */
      gdk_draw_pixmap(canvas,mygc,pete[3],0,0,px,py,96,96);
      gdk_draw_pixmap(canvas,mygc,mbox[1],0,0,500,480-96,96,96);
      pstep[1]++;
      break;
    case 3: /* pete reads */
      gdk_draw_pixmap(canvas,mygc,pete[3],0,0,px,py,96,96);
      
      /* letter */
      gdk_rgb_gc_set_foreground(mygc,0xffffff);
      gdk_draw_rectangle(canvas,mygc,TRUE,px+25,py+56,22,27);
      gdk_rgb_gc_set_foreground(mygc,0);
      gdk_draw_rectangle(canvas,mygc,FALSE,px+25,py+56,22,27);
      gdk_rgb_gc_set_foreground(mygc,0xffffff);

      /* mailbox */
      gdk_draw_pixmap(canvas,mygc,mbox[1],0,0,500,480-96,96,96);

      /* reading ... */
      strcpy(bp,lytaletter);
      p=strtok(bp," ");
      lx=20;
      ly=30;
      for(i=0;i<(s1rstep/2);i++) {
	if (p==NULL) { pstep[1]++; jc=0; break; }
	if (p[0]=='@') { lx=20; ly+=30; p=strtok(NULL," "); continue; }
	if (p[0]=='#') { p=strtok(NULL," "); continue; }
	gdk_draw_string(canvas,avg,mygc,lx,ly,p);
	lx+=gdk_string_width(avg,p)+gdk_string_width(avg," ");
	p=strtok(NULL," ");
      }
      s1rstep++;
      break;
    case 4: /* pete jumps! */

      /* mailbox */
      gdk_draw_pixmap(canvas,mygc,mbox[1],0,0,500,480-96,96,96);

      gdk_draw_pixmap(canvas,mygc,pete[4],0,0,px,py-24*(jc%2),96,96);
      ++jc;
      if (jc>30) {
	pstep[1]++;
	jc=0;
      }
      break;
    case 5: /* pete exclaims! */
      if (px<0) { pstep[1]++; break; }

      gdk_draw_pixmap(canvas,mygc,pete[2+(px/14)%2],0,0,px,py,96,96);
      px-=14;

      gdk_rgb_gc_set_foreground(mygc,0xffff00);
      gdk_draw_string(canvas,avg,mygc,20,20,"- I WON'T LET THE DAMN IRIDIUM");
      gdk_draw_string(canvas,avg,mygc,20,50,"  RUIN MY WEEKEND!");
      
      /* mailbox */
      gdk_draw_pixmap(canvas,mygc,mbox[1],0,0,500,480-96,96,96);
      break;
    case 6:
      prelude=0;
      pstep[1]=0;
      break;
    }

    break;
  }

}
